@@ -30,16 +30,53 @@ from slisemap.utils import (
 
 
 class Slisemap:
-    """Slisemap: combine local explanations with dimensionality reduction.
+    """Slisemap: Combine local explanations with dimensionality reduction.
 
-    __Example Usage__
-    X = np.array([[0.1,0.5,0.7], [0.8,0.9,1], [0.8,0.5,0.3], [0.1,0.2,0.3], [1,2,5], [2,3,4], [2,0,1]])
-    y = np.array([1, 2, 3, 4, 1.5, 1.8, 1.7])
-    sm = Slisemap(X, y, radius=3.5, lasso=1e-4, ridge=2e-4)
-    sm.optimise()
-    sm.plot()
+    Args:
+        X (Union[np.ndarray, torch.Tensor]): Data matrix. Note that the data is assumed to be normalised.
+        y (Union[np.ndarray, torch.Tensor]): Target vector or matrix.
+        radius (float, optional): The radius of the embedding Z. Defaults to 3.5.
+        d (int, optional): The number of embedding dimensions. Defaults to 2.
+        lasso (Optional[float], optional): Lasso regularisation coefficient. Defaults to 0.0.
+        ridge (Optional[float], optional): Ridge regularisation coefficient. Defaults to 0.0.
+        z_norm (float, optional): Z normalisation regularisation coefficient. Defaults to 0.01.
+        intercept (bool, optional): Should an intercept term be added to self.X. Defaults to True.
+        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor], optional): Local model prediction function. Defaults to linear_regression.
+        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor], optional): Local model loss function. Defaults to linear_regression_loss.
+        coefficients (Union[None, int, Callable[[torch.Tensor, torch.Tensor], int]], optional): The number of local model coefficients or a function: `(X,Y)->coefficients`. Defaults to self.X.shape[1] * self.Y.shape[1].
+        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor], optional): Distance function. Defaults to torch.cdist (Euclidean distance).
+        kernel (Callable[[torch.Tensor], torch.Tensor], optional): Kernel function. Defaults to softmax_kernel.
+        B0 (Union[None, np.ndarray, torch.Tensor], optional): Initial value for B (random if None). Defaults to None.
+        Z0 (Union[None, np.ndarray, torch.Tensor], optional): Initial value for Z (PCA if None). Defaults to None.
+        dtype (torch.dtype, optional): Floating type. Defaults to torch.float32.
+        jit (bool, optional): Just-In-Time compile the loss function for increased performance (see `torch.jit.trace` for caveats). Defaults to True.
+        random_state (Optional[int], optional): Set an explicit seed for the random number generator (i.e. `torch.manual_seed`). Defaults to None.
+        cuda (Optional[bool], optional): Use cuda if available (defaults to true if the data is large enough). Defaults to None.
     """
 
+    # Make Python faster and safer by not creating a Slisemap.__dict__
+    __slots__ = (
+        "_X",
+        "_Y",
+        "_Z",
+        "_Z0",
+        "_B",
+        "_B0",
+        "_radius",
+        "_lasso",
+        "_ridge",
+        "_z_norm",
+        "_intercept",
+        "_local_model",
+        "_local_loss",
+        "_loss",
+        "_distance",
+        "_kernel",
+        "_jit",
+        "_random_state",
+        "_rs0",
+    )
+
     def __init__(
         self,
         X: Union[np.ndarray, torch.Tensor],
@@ -56,37 +93,18 @@ class Slisemap:
         local_loss: Callable[
             [torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor
         ] = linear_regression_loss,
-        coefficients: Optional[int] = None,
+        coefficients: Union[
+            None, int, Callable[[torch.Tensor, torch.Tensor], int]
+        ] = None,
         distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = torch.cdist,
         kernel: Callable[[torch.Tensor], torch.Tensor] = softmax_kernel,
         B0: Union[None, np.ndarray, torch.Tensor] = None,
         Z0: Union[None, np.ndarray, torch.Tensor] = None,
         dtype: torch.dtype = torch.float32,
         jit: bool = True,
+        random_state: Optional[int] = None,
         cuda: Optional[bool] = None,
     ):
-        """Create a new Slisemap object.
-
-        Args:
-            X (Union[np.ndarray, torch.Tensor]): Data matrix. Note that the data is assumed to be normalised.
-            y (Union[np.ndarray, torch.Tensor]): Target vector or matrix.
-            radius (float, optional): The radius of the embedding Z. Defaults to 3.5.
-            d (int, optional): The number of embedding dimensions. Defaults to 2.
-            lasso (Optional[float], optional): Lasso regularisation coefficient. Defaults to 0.0.
-            ridge (Optional[float], optional): Ridge regularisation coefficient. Defaults to 0.0.
-            z_norm (float, optional): Z normalisation regularisation coefficient. Defaults to 0.01.
-            intercept (bool, optional): Should an intercept term be added to self.X. Defaults to True.
-            local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor], optional): Local model prediction function. Defaults to linear_regression.
-            local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor], optional): Local model loss function. Defaults to linear_regression_loss.
-            coefficients (Optional[int], optional): The number of local model coefficients. Defaults to self.X.shape[1] * self.Y.shape[1].
-            distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor], optional): Distance function. Defaults to torch.cdist (Euclidean distance).
-            kernel (Callable[[torch.Tensor], torch.Tensor], optional): Kernel function. Defaults to softmax_kernel.
-            B0 (Union[None, np.ndarray, torch.Tensor], optional): Initial value for B (random if None). Defaults to None.
-            Z0 (Union[None, np.ndarray, torch.Tensor], optional): Initial value for Z (PCA if None). Defaults to None.
-            dtype (torch.dtype, optional): Floating type. Defaults to torch.float32.
-            jit (bool, optional): Just-In-Time compile the loss function for increased performance (see `torch.jit.trace` for caveats). Defaults to True.
-            cuda (Optional[bool], optional): Use cuda if available (defaults to true if the data is large enough). Defaults to None.
-        """
         if lasso is None and ridge is None:
             _warn(
                 "Consider using regularisation (lasso/l1 and ridge/l2 regularisation is built-in, via the parameters `lasso` and `ridge`). "
@@ -104,10 +122,11 @@ class Slisemap:
         self.radius = radius
         self._intercept = intercept
         self._jit = jit
+        self._rs0 = random_state
 
         if cuda is None:
-            scale = X.shape[0] ** 2 * (
-                X.shape[1] if coefficients is None else coefficients
+            scale = (
+                X.shape[0] ** 2 * X.shape[1] * (1 if len(y.shape) < 2 else y.shape[1])
             )
             cuda = scale > 1_000_000 and torch.cuda.is_available()
         tensorargs = {
@@ -120,6 +139,8 @@ class Slisemap:
             self._X = torch.cat((self._X, torch.ones_like(self._X[:, :1])), 1)
         n, m = self._X.shape
 
+        self.random_state = random_state
+
         self._Y = torch.as_tensor(y, **tensorargs)
         _assert(
             self._Y.shape[0] == n,
@@ -135,7 +156,7 @@ class Slisemap:
                     "The number of embedding dimensions is larger than the number of data dimensions",
                     Slisemap,
                 )
-                Z0fill = torch.normal(mean=0, std=0.05, size=[n, d - m])
+                Z0fill = torch.zeros(size=[n, d - m], **tensorargs)
                 self._Z0 = torch.cat((self._Z0, Z0fill), 1)
         else:
             self._Z0 = torch.as_tensor(Z0, **tensorargs)
@@ -148,8 +169,11 @@ class Slisemap:
             self._Z0 = self._Z0 * norm
         self._Z = self._Z0.detach().clone()
 
-        coefficients = m * self.p if coefficients is None else coefficients
+        if callable(coefficients):
+            coefficients = coefficients(self._X, self._Y)
         if B0 is None:
+            if coefficients is None:
+                coefficients = m * self.p
             B0 = global_model(
                 X=self._X,
                 Y=self._Y,
@@ -169,6 +193,9 @@ class Slisemap:
                 self._B0 = torch.zeros((n, coefficients), **tensorargs)
         else:
             self._B0 = torch.as_tensor(B0, **tensorargs)
+            if coefficients is None:
+                _assert(len(B0.shape) > 1, "B0 must have more than one dimension")
+                coefficients = B0.shape[1]
             _assert(
                 self._B0.shape == (n, coefficients),
                 f"B0 has the wrong shape: {self._B0.shape} != ({n}, {coefficients})",
@@ -176,9 +203,10 @@ class Slisemap:
         self._B = self._B0.detach().clone()
 
     def restore(self):
-        """Reset B and Z to their initial values (B0 and Z0)."""
+        """Reset B and Z (and random_state) to their initial values B0 and Z0 (and _rs0)."""
         self._Z = self._Z0.clone().detach()
         self._B = self._B0.clone().detach()
+        self.random_state = self._rs0
 
     @property
     def n(self) -> int:
@@ -269,12 +297,12 @@ class Slisemap:
     @property
     def local_model(self) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
         """Local model prediction function. Takes in X[n, m] and B[n, coefficients], and returns Ytilde[n, n, p]."""
-        return self._pred_fn
+        return self._local_model
 
     @local_model.setter
     def local_model(self, value: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]):
         _assert(callable(value), "local_model must be callable")
-        self._pred_fn = value
+        self._local_model = value
         self._loss = None  # invalidate cached loss function
 
     @property
@@ -282,14 +310,14 @@ class Slisemap:
         self,
     ) -> Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]:
         """Local model loss function. Takes in Ytilde[n, n, p], Y[n, p], and B[n, coefficients], and returns L[n, n]"""
-        return self._loss_fn
+        return self._local_loss
 
     @local_loss.setter
     def local_loss(
         self, value: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]
     ):
         _assert(callable(value), "local_loss must be callable")
-        self._loss_fn = value
+        self._local_loss = value
         self._loss = None  # invalidate cached loss function
 
     @property
@@ -324,6 +352,27 @@ class Slisemap:
         self._jit = value
         self._loss = None  # invalidate cached loss function
 
+    def _set_random_state(self, value: Optional[int]):
+        """Set the seed for the random number generator specific for this object (None reverts to the global torch PRNG)."""
+        if value is None:
+            self._random_state = None
+        else:
+            if self.X.device.type == "cpu":
+                self._random_state = torch.random.manual_seed(value)
+            elif self.X.device.type == "cuda":
+                gen = torch.cuda.default_generators[self.X.device.index]
+                self._random_state = gen.manual_seed(value)
+            else:
+                _warn(
+                    Slisemap._set_random_state,
+                    "Unknown device, setting the global seed insted",
+                )
+                torch.random.manual_seed(value)
+                self._random_state = None
+
+    random_state = property(fset=_set_random_state)
+    del _set_random_state
+
     @property
     def X(self) -> torch.Tensor:
         """Data matrix as a `torch.Tensor` (use `Slisemap.get_X()` for options)."""
@@ -351,6 +400,7 @@ class Slisemap:
 
     def cuda(self, **kwargs):
         """Move the tensors to CUDA memory (and run the calculations there).
+        Note that this resets the random state.
 
         Args:
             **kwargs: Optional arguments to `torch.Tensor.cuda`
@@ -360,10 +410,12 @@ class Slisemap:
         self._Y = self._Y.cuda(**kwargs)
         self._Z = self._Z.detach().cuda(**kwargs)
         self._B = self._B.detach().cuda(**kwargs)
+        self.random_state = self._rs0
         self._loss = None  # invalidate cached loss function
 
     def cpu(self, **kwargs):
         """Move the tensors to CPU memory (and run the calculations there).
+        Note that this resets the random state.
 
         Args:
             **kwargs: Optional arguments to `torch.Tensor.cpu`
@@ -373,6 +425,7 @@ class Slisemap:
         self._Y = self._Y.cpu(**kwargs)
         self._Z = self._Z.detach().cpu(**kwargs)
         self._B = self._B.detach().cpu(**kwargs)
+        self.random_state = self._rs0
         self._loss = None  # invalidate cached loss function
 
     def get_loss_fn(
@@ -622,9 +675,11 @@ class Slisemap:
 
         loss_ = self.get_loss_fn()
         loss_fn = lambda: loss_(self._X, self._Y, B, Z)
+        pre_loss = loss_fn().cpu().detach().item()
         LBFGS(loss_fn, [B] if only_B else [Z, B], max_iter=max_iter, **kwargs)
+        post_loss = loss_fn().cpu().detach().item()
 
-        if np.isnan(loss_fn().cpu().detach().item()):
+        if np.isnan(post_loss):
             _warn(
                 "An LBFGS optimisation resulted in `nan` (try strengthening the regularisation or reducing the radius)",
                 Slisemap.lbfgs,
@@ -634,6 +689,9 @@ class Slisemap:
             Z = self._Z.detach().requires_grad_(True)
             B = self._B.detach().requires_grad_(True)
             LBFGS(loss_fn, [B] if only_B else [Z, B], max_iter=1, **kwargs)
+        elif pre_loss <= post_loss:
+            Z = self._Z
+            B = self._B
 
         self._Z = Z.detach()
         self._B = B.detach()
@@ -680,7 +738,7 @@ class Slisemap:
         max_iter: int = 500,
         escape_fn: Callable = escape_neighbourhood,
         verbose: bool = False,
-        noise_scale: float = 1e-4,
+        noise: float = 1e-4,
         **kwargs,
     ) -> float:
         """Optimise Slisemap by alternating between Slisemap.lbfgs and Slisemap.escape until convergence.
@@ -691,7 +749,7 @@ class Slisemap:
             max_iter (int, optional): Maximum number of LBFGS iterations per round. Defaults to 500.
             escape_fn (Callable, optional): Escape function (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
             verbose (bool, optional): Print status messages. Defaults to False.
-            noise_scale (float, optional): Scale of the noise used to avoid loosing dimensions in the embedding after an escape (when using a gradient based optimiser). Defaults to 1e-4.
+            noise (float, optional): Scale of the noise used to avoid disappearing dimensions in the embedding after an escape (gradient based optimisers cannot recover them). Defaults to 1e-4.
             **kwargs: Optional keyword arguments to Slisemap.lbfgs.
 
         Returns:
@@ -707,8 +765,8 @@ class Slisemap:
             loss[1] = self.value()
             if verbose:
                 print(f"Escape {i:2d}: {loss[1]:.2f}")
-            if noise_scale > 0.0:
-                self._Z = torch.normal(self.Z, noise_scale)
+            if noise > 0.0:
+                self._Z = torch.normal(self.Z, noise, generator=self._random_state)
             loss[0] = self.lbfgs(max_iter=max_iter, **kwargs)
             if verbose:
                 print(f"LBFGS  {i+1:2d}: {loss[0]:.2f}")
@@ -719,6 +777,8 @@ class Slisemap:
             self._B = cc.optimal_state._B
         return self.lbfgs(max_iter=max_iter * 2, **kwargs)
 
+    optimize = optimise
+
     def fit_new(
         self,
         Xnew: Union[np.ndarray, torch.Tensor],
@@ -924,20 +984,24 @@ class Slisemap:
         return sm
 
     def _cluster_models(
-        self, clusters: int, B: Optional[np.ndarray] = None
+        self, clusters: int, B: Optional[np.ndarray] = None, random_state: int = 42
     ) -> Tuple[np.ndarray, np.ndarray]:
         """Get cluster labels from k-means on the local model coefficients.
+        This function uses KMeans from scikit-learn.
 
         Args:
             clusters (int): Number of clusters.
             B (Optional[np.ndarray], optional): B matrix. Defaults to self.get_B().
+            random_state (int, optional): random_state for the KMeans clustering. Defaults to 42.
 
         Returns:
             Tuple[np.ndarray, np.ndarray]: vector of cluster labels an matrix of cluster centers.
         """
         from sklearn.cluster import KMeans
 
-        km = KMeans(clusters).fit(B if B is not None else self.get_B())
+        km = KMeans(clusters, random_state=random_state).fit(
+            B if B is not None else self.get_B()
+        )
         influence = np.abs(km.cluster_centers_)
         influence = influence.max(0) + influence.mean(0)
         col = np.argmax(influence)
@@ -951,7 +1015,7 @@ class Slisemap:
         targets: Union[None, str, Sequence[str]] = None,
         clusters: Union[None, int, np.ndarray] = None,
         bars: Union[bool, int] = False,
-        jitter: float = 0.0,
+        jitter: Union[float, np.ndarray] = 0.0,
         B: Optional[np.ndarray] = None,
         Z: Optional[np.ndarray] = None,
         show: bool = True,
@@ -965,7 +1029,7 @@ class Slisemap:
             targets (Union[None, str, Sequence[str]], optional): Target name(s). Defaults to None.
             clusters (Union[None, int, np.ndarray], optional): Can be None (plot individual losses), an int (plot k-means clusters of B), or an array of known cluster id:s. Defaults to None.
             bars (Union[bool, int], optional): If the clusters are from k-means, plot the local models in a bar plot. If `bar` is an int then only plot the most influential variables. Defaults to False.
-            jitter (float, optional): Add random (normal) noise to the scatterplot. Defaults to 0.0.
+            jitter (Union[float, np.ndarray], optional): Add random (normal) noise to the embedding, or a matrix with pre-generated noise matching Z. Defaults to 0.0.
             B (Optional[np.ndarray], optional): Override self.get_B() in the plot. Defaults to None.
             Z (Optional[np.ndarray], optional): Override self.get_Z() in the plot. Defaults to None.
             show (bool, optional): Show the plot. Defaults to True.
@@ -983,7 +1047,9 @@ class Slisemap:
                 "Only the first two dimensions in the embedding are plotted",
                 Slisemap.plot,
             )
-        if jitter > 0:
+        if not isinstance(jitter, float):
+            Z = Z + jitter
+        elif jitter > 0:
             Z = np.random.normal(Z, jitter)
         kwargs.setdefault("figsize", (12, 6))
         fig, (ax1, ax2) = plt.subplots(1, 2, **kwargs)
@@ -1067,10 +1133,11 @@ class Slisemap:
         Y: Union[None, float, np.ndarray, torch.Tensor] = None,
         index: Union[None, int, Sequence[int]] = None,
         title: str = "",
-        jitter: float = 0.0,
+        jitter: Union[float, np.ndarray] = 0.0,
         col_wrap: int = 4,
         selection: bool = True,
         legend_inside: bool = True,
+        Z: Optional[np.ndarray] = None,
         show: bool = True,
         **kwargs,
     ) -> Optional[sns.FacetGrid]:
@@ -1082,10 +1149,11 @@ class Slisemap:
             Y (Union[None, float, np.ndarray, torch.Tensor], optional): Response matrix for the selected data item(s). Defaults to None.
             index (Union[None, int, Sequence[int]], optional): Index/indices of the selected data item(s). Defaults to None.
             title (str, optional): Title of the plot. Defaults to "".
-            jitter (float, optional): Add random (normal) noise to the embedding. Defaults to 0.0.
+            jitter (Union[float, np.ndarray], optional): Add random (normal) noise to the embedding, or a matrix with pre-generated noise matching Z. Defaults to 0.0.
             col_wrap (int, optional): Maximum number of columns. Defaults to 4.
             selection (bool, optional): Mark the selected data item(s), if index is given. Defaults to True.
             legend_inside (bool, optional): Move the legend inside the grid (if there is an empty cell). Defaults to True.
+            Z (Optional[np.ndarray], optional): Override self.get_Z() in the plot. Defaults to None.
             show (bool, optional): Show the plot. Defaults to True.
             **kwargs: Additional arguments to seaborn.relplot.
 
@@ -1094,7 +1162,7 @@ class Slisemap:
         """
         import pandas as pd
 
-        Z = self.get_Z(rotate=True)
+        Z = Z if Z is not None else self.get_Z(rotate=True)
         if Z.shape[1] == 1:
             Z = np.concatenate((Z, np.zeros_like(Z)), 1)
         elif Z.shape[1] > 2:
@@ -1102,7 +1170,9 @@ class Slisemap:
                 "Only the first two dimensions in the embedding are plotted",
                 Slisemap.plot_position,
             )
-        if jitter > 0:
+        if not isinstance(jitter, float):
+            Z = Z + jitter
+        elif jitter > 0:
             Z = np.random.normal(Z, jitter)
         if index is None:
             _assert(
@@ -1196,6 +1266,7 @@ class Slisemap:
         jitter: float = 0.0,
         col_wrap: int = 4,
         legend_inside: bool = True,
+        B: Optional[np.ndarray] = None,
         show: bool = True,
         **kwargs,
     ) -> Optional[sns.FacetGrid]:
@@ -1212,6 +1283,7 @@ class Slisemap:
             jitter (float, optional): Add jitter to the scatterplots. Defaults to 0.0.
             col_wrap (int, optional): Maximum number of columns. Defaults to 4.
             legend_inside (bool, optional): Move the legend inside the grid (if there is an empty cell). Defaults to True.
+            B (Optional[np.ndarray], optional): Override self.get_B() when finding the clusters (only used if clusters is an int). Defaults to None.
             show (bool, optional): Show the plot. Defaults to True.
             **kwargs: Additional arguments to seaborn.relplot.
 
@@ -1237,7 +1309,7 @@ class Slisemap:
 
         if not scatter:
             if isinstance(clusters, int):
-                clusters, _ = self._cluster_models(clusters)
+                clusters, _ = self._cluster_models(clusters, B)
             elif clusters is None:
                 legend_inside = False
             df = pd.concat(
@@ -1253,7 +1325,7 @@ class Slisemap:
                 kwargs.setdefault("common_norm", False)
             kwargs.setdefault("palette", "bright")
             kwargs.setdefault("facet_kws", dict(sharex=False, sharey=False))
-            g = sns.displot(
+            g: sns.FacetGrid = sns.displot(
                 data=df,
                 x="Value",
                 hue=None if clusters is None else "Cluster",
@@ -1261,9 +1333,8 @@ class Slisemap:
                 col_wrap=col_wrap,
                 **kwargs,
             )
-            g.set_titles("")
-            for ax, n in zip(g.axes.flat, g.col_names):
-                ax.set_xlabel(n)
+            g.set_titles("{col_name}")
+            g.set_xlabels("")
         else:
             Z = self.get_Z(rotate=True)
             if Z.shape[1] == 1:
